<?php

/**
 * @file
 * Implements guild tabard image creation.
 */

/**
 * Partial path to Battle.net tabard base images.
 */
define('BLIZZARDAPI_GUILD_TABARDS_URL', '/images/guild/tabards/');
/**
 * Maximum tabard image size.
 */
define('BLIZZARDAPI_TABARD_MAX_SIZE', 240);
/**
 * Minimum tabard image size.
 */
define('BLIZZARDAPI_TABARD_MIN_SIZE', 16);

/**
 * Get the path to a tabard base image. If there is no local copy of the image,
 * it will be downloaded from the specified region's Battle.net site.
 *
 * @param string $filename
 *   The name of the image file to look for.
 * @param string $region
 *   (optional) The armory region to download any necessary base images from.
 * 
 * @return string The path to the tabard image or FALSE if an error occurred.
 */
function blizzardapi_get_tabard_image($filename, $region = 'us') {
  $paths['blizzardapi'] = file_create_path('blizzardapi');
  $paths['tabards'] = $paths['blizzardapi'] . '/tabards';
  $paths['base'] = $paths['tabards'] . '/base';
  foreach ($paths as $path) {
    file_check_directory($path, FILE_CREATE_DIRECTORY);
  }
  
  $image = $paths['base'] . '/' . $filename;
  if (!file_exists($image)) {
    $host = ($region == 'cn') ? BLIZZARDAPI_URL_BASE_CHINA : sprintf(BLIZZARDAPI_URL_BASE, $region);
    $url = 'http:' . $host . BLIZZARDAPI_STATIC_URL . BLIZZARDAPI_GUILD_TABARDS_URL . $filename;
    
    $response = drupal_http_request($url);
    switch ($response->code) {
      case 200: // OK
      case 302: // Found
      case 307: // Temporary Redirect
        if (!file_save_data($response->data, $image, FILE_EXISTS_ERROR)) {
          watchdog('blizzardapi', 'Unable to save tabard image to %file.', array('%file' => $image), WATCHDOG_ERROR);
          $image = FALSE;
        }
        else {
          // This does count as a webserver-generated file right?
          @chmod($image, 0664);
          
          watchdog('blizzardapi', 'Downloaded base tabard image to %file.', array('%file' => $image), WATCHDOG_NOTICE);
        }
        break;
      default:
        watchdog('blizzardapi', 'Unable to download tabard image from %url. Request response: [%code] %error.',
          array('%url' => $url, '%code' => $response->code, '%error' => $response->error), WATCHDOG_ERROR);
        $image = FALSE;
        break;
    }
  }
  
  return $image;
}

/**
 * Generate a guild tabard image (in PNG format).
 * 
 * >>> This function REQUIRES the transliteration module. <<<
 * 
 * Files will be saved at files/blizzardapi/tabards/render/{region}-{realm}/{name}.png.
 * If the file already exists, it will be overwritten.
 *
 * @param array $guild
 *   (optional) An array containing guild information as returned by the API
 *   resource. If not provided a default tabard will be generated.
 * @param int $tabard_size
 *   (optional) The size of the tabard image to create. The maximum size is
 *   defined by BLIZZARDAPI_TABARD_MAX_SIZE which is 240px. The minimum size is
 *   defined by BLIZZARDAPI_TABARD_MIN_SIZE which is 16px.
 * @param string $region
 *   (optional) The armory region to download any necessary base images from.
 *   Additional realm information will also be retrieved from this region.
 * @param string $language
 *   (optional) The transliteration source language, defaults to the site language.
 *
 * @return The path to the generated tabard image or FALSE if there was an error.
 */
function blizzardapi_create_tabard($guild = array(), $tabard_size = 240, $region = 'us', $language = '') {
  if ($tabard_size < BLIZZARDAPI_TABARD_MIN_SIZE || $tabard_size > BLIZZARDAPI_TABARD_MAX_SIZE) {
    return FALSE;
  }
  
  $toolkits = image_get_available_toolkits();
  if (!isset($toolkits['gd'])) {
    return FALSE;
  }
  
  // Note that the order of each image is important.
  $images = array(
    'ring' => array(
      'file' => 'ring-' . ($guild['side'] == 0 ? 'alliance' : 'horde') . '.png',
      'position' => array('x' => 0, 'y' => 0)
    ),
    'shadow' => array(
      'file' => 'shadow_00.png',
      'position' => array('x' => 18, 'y' => 27)
    ),
    'background' => array(
      'file' => 'bg_00.png',
      'position' => array('x' => 18, 'y' => 27),
      'color' => $guild['emblem']['backgroundColor']
    ),
    'overlay' => array(
      'file' => 'overlay_00.png',
      'position' => array('x' => 18, 'y' => 27)
    ),
    'border' => array(
      'file' => sprintf('border_%02d.png', $guild['emblem']['border']),
      'position' => array('x' => 31, 'y' => 40),
      'color' => $guild['emblem']['borderColor']
    ),
    'icon' => array(
      'file' => sprintf('emblem_%02d.png', $guild['emblem']['icon']),
      'position' => array('x' => 33, 'y' => 57),
      'color' => $guild['emblem']['iconColor']
    ),
    'hooks' => array(
      'file' => 'hooks.png',
      'position' => array('x' => 18, 'y' => 27)
    )
  );
  
  if (empty($guild['realm'])) {
    $realm = $name = 'default';
    unset($images['border'], $images['icon']);
  }
  else {
    if ($guild['emblem']['border'] < 0 || $guild['emblem']['icon'] < 0) {
      unset($images['border'], $images['icon']);
    }
    
    // Until PHP 6, passing UTF-8 strings to filesystem functions will produce
    // incorrect results, thereby requiring transliteration.
    // https://bugs.php.net/bug.php?id=46990
    if (!function_exists('transliteration_clean_filename')) {
      return FALSE;
    }
    
    if (empty($language)) {
      $language = language_default('language');
    }
    
    $slug = transliteration_clean_filename(blizzardapi_realm_slug($guild['realm'], $region), $language);
    $name = transliteration_clean_filename($guild['name'], $language);
    $realm = $region . '-' . $slug;
  }
  
  // Prepare the target location for the generated file.
  $paths['blizzardapi'] = file_create_path('blizzardapi');
  $paths['tabards'] = $paths['blizzardapi'] . '/tabards';
  $paths['render'] = $paths['tabards'] . '/render';
  $paths['realm'] = $paths['render'] . '/' . $realm;
  foreach ($paths as $path) {
    if (!file_check_directory($path, FILE_CREATE_DIRECTORY)) {
      return FALSE;
    }
  }
  
  $filename = $paths['realm'] . '/' . $name . '.png';
  
  // Prepare the target buffer.
  $target = imagecreatetruecolor($tabard_size, $tabard_size);
  imagesavealpha($target, TRUE);
  $transparent = imagecolorallocatealpha($target, 0, 0, 0, 127);
  imagefill($target, 0, 0, $transparent);
  
  // Add each tabard image.
  foreach ($images as $fragment => $info) {
    $src = blizzardapi_get_tabard_image($info['file']);
    
    // Unable to download base image, cleanup and quit.
    if ($src == FALSE) {
      imagedestroy($target);
      return FALSE;
    }
    
    $color = isset($info['color']) ? $info['color'] : NULL;
    _blizzardapi_tabard_copy($target, $src, $info['position'], array(), $color);
  }
  
  // Save the tabard and cleanup the buffer.
  imagepng($target, $filename);
  imagedestroy($target);
  
  // Set standard file permissions for webserver-generated files.
  @chmod(realpath($filename), 0664);
  
  return $filename;
}

/**
 * Converts a hexadecimal color to its aRGB components.
 */
function _blizzardapi_unpack_color($hex, $normalize = FALSE) {
  $rgb = array();
  
  if (strlen($hex) == 8) {
    // Split the segments.
    $alpha = substr($hex, 0, 2);
    $hex = substr($hex, 2);
    
    // GD functions expect a 7-bit alpha (0-127).
    $rgb['a'] = abs(hexdec($alpha) - 255) >> 1;
  }
  
  $c = hexdec($hex);
  
  for ($i = 16; $i >= 0; $i -= 8) {
    $out[] = (($c >> $i) & 0xFF) / ($normalize ? 255 : 1);
  }
  
  // added = color / intensity + color * blend
  // scale = color / 255 + blend
  // color = color * scale + added
  
  // color = color * (color / 255 + blend) + (color / intensity + color * blend)
  
  //$intensity = 19.0;
  //$blend = 1.0 / 3.0;
  //$out[0] = $out[0] * ($out[0] / 255 + $blend) + ($out[0] / $intensity + $out[0] * $blend);
  //$out[1] = $out[1] * ($out[1] / 255 + $blend) + ($out[1] / $intensity + $out[1] * $blend);
  //$out[2] = $out[2] * ($out[2] / 255 + $blend) + ($out[2] / $intensity + $out[2] * $blend);
  
  // normal:    {134,15,154} {15,20,21} {16,21,23}
  // colorized: {166,11,203} {11,15,16} {12,16,18}
  
  $rgb += array(
    'r' => $out[0],
    'g' => $out[1],
    'b' => $out[2]
  );
  
  return $rgb;
}

/**
 * Copies a rectangular portion of one tabard image to another image and applies
 * an overlay color onto the image if specified.
 *
 * @param resource $target
 *   An image resource, returned by one of the image creation functions, such as
 *   imagecreatetruecolor().
 * @param string $src
 *   Path of the image to overlay on the target. This must be a PNG image.
 * @param array $target_point
 *   Location to place the source image on the destination. Coordinates refer to
 *   the upper left corner.
 * @param array $src_rect
 *   (optional) An array defining the area of the source image to copy. The array
 *   should contain the following keys: {x, y, width, height}.
 * @param mixed $overlay_color
 *   (optional) The color to overlay on the source image. Defaults to none.
 *
 *   The following values are supported:
 *   - A hexadecimal string representing the color.
 *   - An array containing the aRGB values of the color.
 *   - A color identifier as returned by imagecolorallocate().
 */
function _blizzardapi_tabard_copy($target, $src, $target_point, $src_rect = array(), $overlay_color = NULL) {
  $image = imagecreatefrompng($src);
  
  // Hexadecimal -> RGB
  if (is_string($overlay_color)) {
    $overlay_color = _blizzardapi_unpack_color($overlay_color);
  }
  // RGB -> GD color
  if (is_array($overlay_color)) {
    $overlay_color = imagecolorallocate($image, $overlay_color['r'], $overlay_color['g'], $overlay_color['b']);
  }
  
  $width = imagesx($image);
  $height = imagesy($image);
  
  if (!is_null($overlay_color)) {
    imagelayereffect($image, IMG_EFFECT_OVERLAY);
    imagefilledrectangle($image, 0, 0, $width, $height, $overlay_color);
  }
  
  // Scale the source image onto the target image at the given point.
  $size = imagesx($target);
  $target_rect = array(
    'x' => $size * $target_point['x'] / BLIZZARDAPI_TABARD_MAX_SIZE,
    'y' => $size * $target_point['y'] / BLIZZARDAPI_TABARD_MAX_SIZE,
    'width' => $size * $width / BLIZZARDAPI_TABARD_MAX_SIZE,
    'height' => $size * $height / BLIZZARDAPI_TABARD_MAX_SIZE
  );
  
  // Default to copying the entire source image.
  if (empty($src_rect)) {
    $src_rect = array(
      'x' => 0,
      'y' => 0,
      'width' => $width,
      'height' => $height
    );
  }
  
  // Seriously, 10 parameters!
  imagecopyresampled(
    $target, $image,
    $target_rect['x'], $target_rect['y'],
    $src_rect['x'], $src_rect['y'],
    $target_rect['width'], $target_rect['height'],
    $src_rect['width'], $src_rect['height']
  );
  
  imagedestroy($image);
}

/**
 * Determine if a tabard has already been generated for the specified guild.
 */
function blizzardapi_tabard_exists($realm, $name, $region = 'us', $language = '') {
  if (!function_exists('transliteration_clean_filename')) {
    return FALSE;
  }
  
  if (empty($language)) {
    $language = language_default('language');
  }
  
  $base = file_create_path('blizzardapi') . '/tabards/render';
  $slug = transliteration_clean_filename(blizzardapi_realm_slug($realm, $region), $language);
  $name = transliteration_clean_filename($name, $language);
  
  $tabard = "$base/$region-$slug/$name.png";
  
  return file_exists($tabard);
}
